@@ -1,6 +1,6 @@
 package UvT::Email::Template;
 
-# $Id: Template.pm 39318 2013-05-02 13:49:32Z wsl $
+# $Id: Template.pm 47785 2018-09-13 12:29:25Z wsl $
 # $URL: https://svn.uvt.nl/its-id/trunk/sources/uvt-email-template/lib/UvT/Email/Template.pm $
 
 # formatteer een XML-bericht tot een multipart/alternative MIME
@@ -14,8 +14,10 @@ use Clarity -self;
 use XML::LibXML;
 use XML::LibXSLT;
 use Email::MIME::Creator;
+use MIME::Types;
+use Image::Size ();
 use Text::Tabs qw(expand);
-use Scalar::Util qw(weaken reftype);
+use Scalar::Util qw(weaken reftype refaddr);
 
 {
 	# perl 5.12+ kan dit met each() maar zit niet in squeeze :\
@@ -23,26 +25,27 @@ use Scalar::Util qw(weaken reftype);
 	my %array_each;
 	my %array_each_i;
 
-	sub array_each(+) {
+	sub array_each(\@) {
 		my $arr = shift;
 		confess "Not an array or arrayref"
 			unless reftype($arr) eq 'ARRAY';
-		unless(defined $array_each{$arr}) {
+		my $addr = refaddr($arr);
+		unless(defined $array_each{$addr}) {
 			while(my ($k, $v) = each(%array_each)) {
 				unless(defined $v) {
 					delete $array_each{$k};
 					delete $array_each_i{$k};
 				}
 			}
-			$array_each{$arr} = $arr;
-			weaken($array_each{$arr});
-			$array_each_i{$arr} = 0;
+			$array_each{$addr} = $arr;
+			weaken($array_each{$addr});
+			$array_each_i{$addr} = 0;
 		}
-		my $i = $array_each_i{$arr}++;
+		my $i = $array_each_i{$addr}++;
 		return wantarray ? ($i, $arr->[$i]) : $i
 			if $i < @$arr;
-		delete $array_each{$arr};
-		delete $array_each_i{$arr};
+		delete $array_each{$addr};
+		delete $array_each_i{$addr};
 		return;
 	}
 }
@@ -67,7 +70,7 @@ sub wrap {
 	my ($init, $cont, $text) = @_;
 	my @lines = split(/(\n)/, $text);
 	my $pfx = $init;
-	my $res;
+	my $res = '';
 	while(@lines) {
 		my $line = shift(@lines);
 		my $eol = shift(@lines) // '';
@@ -129,7 +132,7 @@ sub format_strong {
 
 sub format_em {
 	my $text = $self->nodes_to_text(shift->childNodes);
-#		return "/$text/";
+#	return "/$text/";
 	$text =~ s/([^_\n])/$1\010_/g;
 	return $text;
 }
@@ -174,7 +177,7 @@ sub format_td {
 
 sub format_table {
 	my $node = shift;
-	die "Table text must be inside th/td element\n"
+	die "table text must be inside th/td element\n"
 		if $node->exists("text()[normalize-space()!='']|*/text()[normalize-space()!='']");
 	my @r;
 	foreach my $r ($node->childNodes) {
@@ -184,6 +187,7 @@ sub format_table {
 			unless $name eq 'tr';
 		my @c;
 		foreach my $c ($r->childNodes) {
+			next unless $r->nodeType == XML_ELEMENT_NODE;
 			my $name = $c->nodeName // next;
 			die "<tr> element must only contain <th> or <td> elements\n"
 				unless $name eq 'th' || $name eq 'td';
@@ -198,7 +202,7 @@ sub list_to_text {
 	my ($fmt, $node) = @_;
 	my $width = length(sprintf($fmt, 0));
 	my $indent = " "x$width;
-	my $index;
+	my $index = 0;
 	my $text = '';
 	foreach my $item ($node->findnodes('li')) {
 		$index++;
@@ -214,13 +218,13 @@ sub list_to_text {
 }
 
 sub nodes_to_text {
-	my $text;
+	my $text = '';
 	foreach my $node (@_) {
 		my $type = $node->nodeType;
 		if($type == XML_ELEMENT_NODE) {
 			my $name = $node->nodeName;
 			my $h = "format_$name";
-			die "Unknown element type '$name' encountered\n"
+			die "unknown element type '$name' encountered\n"
 				unless $self->can($h);
 			$text .= $self->$h($node);
 		} elsif($type == XML_TEXT_NODE || $type == XML_CDATA_SECTION_NODE) {
@@ -230,7 +234,7 @@ sub nodes_to_text {
 		} elsif($type == XML_ATTRIBUTE_NODE || $type == XML_COMMENT_NODE) {
 			# ignore
 		} else {
-			die "Unknown node type $type encountered\n";
+			die "unknown node type $type encountered\n";
 		}
 	}
 	return $text;
@@ -324,8 +328,8 @@ sub build_table {
 			my $text = $self->nodes_to_par($col);
 			$text =~ s/\s+$//s;
 			push @cells, $text;
-			my $h = split(/\n/, $text);
-			$height = $h if $height < $h;
+			my @h = split(/\n/, $text);
+			$height = @h if $height < @h;
 		}
 		my $rowres = "│\n"x$height;
 		while(my ($i, $cell) = array_each(@cells)) {
@@ -359,15 +363,27 @@ sub kop_to_text {
 
 sub doc_to_text {
 	my $doc = shift;
-	my @pars = (
-		"(English version below)",
-		$self->kop_to_text($doc->findvalue('mailtje/nederlands/kop')),
-		(map { $self->par_to_text($_) } $doc->findnodes('mailtje/nederlands/par|mailtje/allebei/par')),
-		('─'x$columns),
-		$self->kop_to_text($doc->findvalue('mailtje/engels/kop')),
-		(map { $self->par_to_text($_) } $doc->findnodes('mailtje/engels/par|mailtje/allebei/par')),
-		"-- \nTilburg University <www.tilburguniversity.edu> – Warandelaan 2, 5037 AB Tilburg\n\nroutebeschrijving: <http://www.tilburguniversity.edu/nl/contact/>\nroute description: <http://www.tilburguniversity.edu/contact/>"
-	);
+	my @pars;
+
+	my $have_nederlands = $doc->exists('mailtje/nederlands');
+	my $have_engels = $doc->exists('mailtje/engels');
+
+	if($have_nederlands) {
+		push @pars, "(English version below)" if $have_engels;
+		push @pars,
+			$self->kop_to_text($doc->findvalue('mailtje/nederlands/kop')),
+			(map { $self->par_to_text($_) } $doc->findnodes('mailtje/nederlands/par|mailtje/allebei/par'));
+	}
+	if($have_engels) {
+		push @pars, ('─'x$columns) if $have_nederlands;
+		push @pars,
+			$self->kop_to_text($doc->findvalue('mailtje/engels/kop')),
+			(map { $self->par_to_text($_) } $doc->findnodes('mailtje/engels/par|mailtje/allebei/par'));
+	}
+
+	push @pars,
+		"-- \nTilburg University <www.tilburguniversity.edu> – Warandelaan 2, 5037 AB Tilburg\n\nroutebeschrijving: <http://www.tilburguniversity.edu/nl/contact/>\nroute description: <http://www.tilburguniversity.edu/contact/>";
+
 	my $text = join("\n", map { "$_\n" } @pars);
 	utf8::encode($text);
 	return $text;
@@ -390,13 +406,6 @@ field xml => sub {
 
 field xslt => sub { new XML::LibXSLT };
 
-field mimetypes => {
-	gif => 'image/gif',
-	png => 'image/png',
-	jpg => 'image/jpeg',
-	jpeg => 'image/jpeg',
-};
-
 sub htmltemplate {
 	return $self->{htmltemplate} unless @_;
 	my $t = shift;
@@ -405,7 +414,7 @@ sub htmltemplate {
 	return $self->{htmltemplate} = $t;
 }
 
-field imagepath => ['.'];
+field imagepath => '.';
 
 sub transformer {
 	return $self->{transformer} unless @_;
@@ -415,9 +424,13 @@ sub transformer {
 	return $self->{transformer} = $t;
 }
 
+field mimetypes => sub { new MIME::Types };
+
 sub format {
 	my $doc = shift;
 
+	my $images = ref $_[0] ? shift : {};
+
 	$doc = $self->xml->parse_file($doc)
 		unless ref $doc;
 
@@ -438,38 +451,94 @@ sub format {
 	$xpc->registerNs(html => 'http://www.w3.org/1999/xhtml');
 
 	my $imagepath = $self->imagepath;
-	my $mimetypes = $self->mimetypes;
+	$imagepath = [$imagepath] unless ref $imagepath;
+
+	my $mimetypes;
+
+	my %attachedimages;
 
 	foreach my $img ($xpc->findnodes('//html:img', $html_dom)) {
 		my $src = $img->getAttribute('src')
 			// die "img element has no 'src' attribute\n";
 		next if $src =~ /^[a-z0-9_-]+:/i;
-		die "Refusing to open image file with funny name ($src)\n"
-			if $src !~ /^[a-z0-9]+(?:[_-][a-z0-9]+)*\.(gif|png|jpe?g)$/;
-		my $type = $mimetypes->{lc($1)};
-		die "Unknown image extension '$1'\n"
-			unless defined $type;
-		my $path;
-		foreach my $p (@$imagepath) {
-			next unless -r "$p/$src";
-			$path = "$p/$src";
-			last;
+
+		my ($id, $width, $height);
+
+		if(exists $attachedimages{$src}) {
+			($id, $width, $height) = @{$attachedimages{$src}};
+		} else {
+			my $data;
+			my $type;
+			my $name = $src;
+
+			if(exists $images->{$src}) {
+				my $spec = $images->{$src};
+				if(ref $spec) {
+					$type = $spec->{type};
+					$data = $spec->{data};
+				}
+			} else {
+				die "refusing to open image file with funny name ($src)\n"
+					unless $src =~ /^[a-z0-9]+(?:[_\/.-][a-z0-9]+)*$/;
+
+				my $path;
+				foreach my $p (@$imagepath) {
+					next unless -f "$p/$src";
+					$path = "$p/$src";
+					last;
+				}
+				die "image '$src' not found in imagepath [@$imagepath]\n"
+					unless defined $path;
+
+				$name = $path;
+				$data = $self->load_file($path);
+			}
+
+			($width, $height, my $imgtype) = Image::Size::imgsize(\$data);
+
+			if(defined $type) {
+				die "file '$name' ($type) is not an image\n"
+					unless $type =~ m{^image/};
+				if(defined $imgtype) {
+					$mimetypes //= $self->mimetypes;
+					my $mimetype = $mimetypes->mimeTypeOf($imgtype);
+					unless(defined $mimetype && $mimetype->type eq $type) {
+						# something is wonky, play safe
+						undef $width;
+						undef $height;
+					}
+				}
+			} else {
+				$mimetypes //= $self->mimetypes;
+				my $mimetype = $mimetypes->mimeTypeOf($imgtype // $src);
+				die "unable to determine mime type for '$name'\n"
+					unless defined $mimetype;
+				$type = $mimetype->type;
+				die "file '$name' ($type) is not an image\n"
+					unless $mimetype->mediaType eq 'image';
+			}
+
+			$id = @attachments + 1;
+			push @attachments, Email::MIME->create(
+				attributes => {
+					content_type => $type,
+					encoding => 'base64',
+					disposition => 'inline',
+					name => $src,
+					filename => $src,
+				},
+				body => $data,
+				header_str => ['Content-ID' => "<$id>"]
+			);
+
+			$attachedimages{$src} = [$id, $width, $height];
 		}
-		die "Image '$src' not found in imagepath [@$imagepath]\n"
-			unless defined $path;
-		my $id = @attachments + 1;
+
 		$img->setAttribute(src => "cid:$id");
-		push @attachments, Email::MIME->create(
-			attributes => {
-				content_type => $type,
-				encoding => 'base64',
-				disposition => 'inline',
-				name => $src,
-				filename => $src,
-			},
-			body => $self->load_file($path),
-			header_str => ['Content-ID' => "<$id>"]
-		);
+		unless(defined $img->getAttribute('width') || defined $img->getAttribute('height')) {
+			$img->setAttribute(width => $width) if defined $width;
+			$img->setAttribute(height => $height) if defined $height;
+		}
 	}
 
 	my $html = $html_dom->toStringHTML;
@@ -508,7 +577,7 @@ UvT::Email::Format - UvT style email formatting
 =head1 INTRODUCTION
 
 This class can create ready to use e-mails formatted using XSL templates
-(for HTML) or a built-in XML converter (for text). The resulting MIME
+(for HTML) and a built-in XML converter (for text). The resulting MIME
 structure is a multipart/alternative structure containing both the HTML
 and text versions.
 
@@ -535,17 +604,17 @@ This field contains the XML parser options that are used to parse any XML.
 It's a hash that you can either replace or modify. See XML::LibXML::Parser
 for details.
 
-=head2 mimetypes
-
-A hash that maps file extensions to MIME types. You can replace or modify
-it.
-
 =head2 transformer
 
 An optional stylesheet that will be applied to your XML document when you
 call $e->format(). Can be either a filename or a completely prepared
 XML::LibXSLT::Stylesheet object.
 
+=head2 imagepath
+
+This field holds an array of strings that is used as the search path for
+images. The default value is C<['.']>.
+
 =head2 format($doc, Header => "value", ...)
 
 The workhorse of this class; it takes an XML filename or completely
@@ -555,6 +624,95 @@ value of this method is a string that forms a complete RFC2822 e-mail.
 Additional parameters are interpreted as extra headers that will be added
 to the resulting e-mail message (with appropriate encoding).
 
+=head1 INPUT
+
+The input for the format() method is an XML file with the following structure:
+
+  <mailtje>
+    <disclaimer/>
+
+    <nederlands>
+      <kop>text</kop>
+      <par>html</par>
+      <par>html</par>
+      ...
+    </nederlands>
+    <engels>
+      <kop>text</kop>
+      <par>html</par>
+      <par>html</par>
+      ...
+    </engels>
+
+    <allebei>
+      <par>html</par>
+      <par>html</par>
+      ...
+    </allebei>
+
+    <nederlands>
+      <par>html</par>
+      <par>html</par>
+      ...
+    </nederlands>
+    <engels>
+      <par>html</par>
+      <par>html</par>
+      ...
+    </engels>
+    ...
+  </mailtje>
+
+The root tag of this XML document must be "mailtje". The document must contain
+at least one "nederlands" or "engels" element.
+
+The "allebei" elements are optional. They describe a text part that will
+appear in both language sections in the output. In the HTML output this
+part will be centered and span both columns.
+
+The first "nederlands" and the first "engels" elements must have exactly
+one "kop" element as the first member. No other elements may contain a
+"kop" element.
+
+The empty <disclaimer/> tag is optional and indicates whether you want a
+disclaimer about automated translations in the HTML version.
+
+The allowed html elements and attributes are:
+
+=over
+
+=item strong
+
+=item em
+
+=item ul
+
+=item ol
+
+=item li
+
+=item a
+
+(the "href" attribute is required)
+
+=item img
+
+(the "src" and "alt" attributes are required)
+
+=item br
+
+=item table
+
+=item tr
+
+=item th
+
+=item td
+
+=back
+
+Unless noted otherwise, no attributes are allowed.
+
 =head1 AUTHOR
 
 Wessel Dankers <wsl@uvt.nl>
